package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_ALASKA_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_ALASKA_INTERFACE_ADJUSTED;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_ALASKA_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_ALASKA_SLAB_ADJUSTED;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_INTERFACE_ADJUSTED;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_INTERFACE_ADJUSTED_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_INTERFACE_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_SLAB_ADJUSTED;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_SLAB_ADJUSTED_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_CASCADIA_SLAB_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_GLOBAL_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_GLOBAL_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGV;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.ALASKA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.CASCADIA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.GLOBAL;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.SLAB;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.min;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the Abrahamson &amp; Gülerce (2020) ground motion model for
 * subduction regions developed as part of the <a
 * href="https://www.risksciences.ucla.edu/nhr3/nga-subduction">NGA
 * Subduction</a> project.
 *
 * <p><b>Implementation notes:</b>
 *
 * <ul><li>Implementation currently supports a global median model, as well as
 * Alaska and Cascadia regional models. Following the Earthquake Spectra version
 * this implementation provides epistemic uncertainty on the median ground
 * motion for the Alaska and Cascadia models, but not the global model.</li>
 *
 * <li>Coefficients for 0.01s are used for PGA.</li>
 *
 * <li>Unlike BcHydro, this model does not include a forearc-backarc term.</li>
 *
 * <li>Mean model does not include aftershock term.</li>
 *
 * <li>{@code zTor} is interpreted as hypocentral depth and is only used for
 * slab events; it is limited to 120 km, consistent with other subduction
 * GMMs.</li>
 *
 * <li>Basin scaling only amplifies long period ground motion when z2.5 is
 * greater than the reference value computed from Vs30. If z2.5 is supplied, the
 * Cascadia basin term is used for all implementations.</li>
 *
 * <li>The aleatory variability model only includes the linear φ1 model
 * appropriate for the global, Alaska and Cascadia models.</li>
 *
 * <li>PGV supported via {@link UsgsPgvSupport}.</li></ul>
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Reference:</b> Abrahamson, N.A., and Gülerce, Z., 2022, Summary of the
 * Abrahamson and Gulerce NGA-SUB ground-motion model for subduction
 * earthquakes, Earthquake Spectra, v. 38, n. 4, p. 2638–2681, doi: <a
 * href="https://doi.org/10.1177/87552930221114374">10.1177/87552930221114374</a>.
 *
 * <p><b>Reference:</b> Abrahamson, N., and Gülerce, Z., 2020, Regionalized
 * ground-motion models for subduction earthquakes based on the NGA-SUB
 * database. Report no. 2020/25. Berkeley, CA: PEER, 219 pp.
 *
 * <p><b>Component:</b> average horizontal (RotD50)
 *
 * @author U.S. Geological Survey
 */
public abstract class AbrahamsonGulerce_2020 implements GroundMotionModel {

  static final String NAME = "Abrahamson & Gülerce (2020)";

  static final Constraints CONSTRAINTS_INTERFACE = Constraints.builder()
      .set(MW, Range.closed(6.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0)) // check report
      .set(ZTOR, Earthquakes.INTERFACE_DEPTH_RANGE)
      .set(VS30, Range.closed(150.0, 1000.0))
      .build();

  static final Constraints CONSTRAINTS_SLAB = Constraints.builder()
      .set(MW, Range.closed(5.0, 8.0))
      .set(RRUP, Range.closed(0.0, 1000.0)) // check report
      .set(ZTOR, Earthquakes.SLAB_DEPTH_RANGE)
      .set(VS30, Range.closed(150.0, 1000.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("AG20.csv");
  static final CoefficientContainer COEFFS_AK_BIAS =
      new CoefficientContainer("nga-sub-ak-interface-bias-corr.csv");

  private static final double C1S = 7.5;
  private static final double C4 = 10.0;
  private static final double C = 1.88;
  private static final double N = 1.18;

  private static final double VSS_MAX = 1000.0;
  private static final double VS30_ROCK = 1000.0;

  private static final double D0 = 0.47;

  private static final class Coefficients {

    final Imt imt;
    final double c1i, c1s;
    final double a1, a2, a3, a4, a5, a6, a8, a9, a10, a11, a12, a13, a14, a39, a45;
    final double d1, d2, ρB, ρW;
    final double vlin, b;
    final double a1Adj;
    final double ε; // epistemic uncertainty in mean

    Coefficients(Imt imt, CoefficientContainer cc, SubductionZone zone) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);

      /* Regional adjustments (Global) */
      double a1r = coeffs.get("a1");
      double a1AdjRegional = 0.0;
      double a6r = 0.0;
      double a12r = 0.0;
      double c1sr = 7.5;

      /* Global epistemic uncertainty (NAA pers. comm. 5/11/23) */
      double εr = 0.26;
      if (imt.isSA() && imt.period() > 3.0) {
        εr += 0.15 * log(imt.period() / 3.0);
      }

      if (zone == ALASKA) {
        a1r = coeffs.get("a31");
        a1AdjRegional = coeffs.get("cAk");
        a6r = coeffs.get("a24");
        a12r = coeffs.get("a17");
        c1sr = 7.9;
        εr = 0.21;
      } else if (zone == CASCADIA) {
        a1r = coeffs.get("a32");
        a1AdjRegional = coeffs.get("cCasc");
        a6r = coeffs.get("a25");
        a12r = coeffs.get("a18");
        c1sr = 7.1;
        εr = 0.27;
      }

      /* M-scaling break points */
      c1i = coeffs.get("c1i");
      c1s = c1sr;

      ε = εr;

      a1 = a1r;
      a1Adj = a1AdjRegional;
      a2 = coeffs.get("a2");
      a3 = coeffs.get("a3");
      a4 = coeffs.get("a4");
      a5 = coeffs.get("a5");
      a6 = coeffs.get("a6") + a6r;
      a8 = coeffs.get("a8");
      a9 = coeffs.get("a9");
      a10 = coeffs.get("a10");
      a11 = coeffs.get("a11");
      a12 = coeffs.get("a12") + a12r;
      a13 = coeffs.get("a13");
      a14 = coeffs.get("a14");
      a39 = coeffs.get("a39");
      a45 = coeffs.get("a45");
      d1 = coeffs.get("d1");
      d2 = coeffs.get("d2");
      ρB = coeffs.get("rhoB");
      ρW = coeffs.get("rhoW");
      vlin = coeffs.get("vlin");
      b = coeffs.get("b");
    }
  }

  /*
   * Average bias-correction for NGA-Subduction interface models in Alaska nshm.
   */
  private static final class CoefficientsAkBias {
    final double correction;
    final double correctionPGA;

    CoefficientsAkBias(Imt imt, CoefficientContainer cc) {
      correction = cc.get(imt).get("bias_ak");
      correctionPGA = cc.get(PGA).get("bias_ak");
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;
  private final CoefficientsAkBias coeffAkBias;
  private final Gmm subtype;

  AbrahamsonGulerce_2020(Imt imt, Gmm subtype, SubductionZone zone) {
    this.subtype = subtype;
    coeffs = new Coefficients(imt, COEFFS, zone);
    coeffsPGA = new Coefficients(PGA, COEFFS, zone);
    coeffAkBias = new CoefficientsAkBias(imt, COEFFS_AK_BIAS);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  /* Intraslab (true) or interface (false) model. */
  boolean slab() {
    return false;
  }

  /* Use USGS deep basin, long period amplification model. */
  boolean basin() {
    return false;
  }

  /* Use 'adjusted' models; only applicable to Cascadia and Alaska). */
  boolean adjusted() {
    return false;
  }

  /* Apply USGS bias adjustment for Alaska */
  boolean akBiasAdjusted() {
    return false;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {

    if (coeffs.imt == PGV) {
      GroundMotion pgv = UsgsPgvSupport.calcAB20Pgv(subtype, in);
      return GroundMotions.createTree(pgv.mean(), pgv.sigma());
    }

    double pgaRock = 0.0;
    if (in.vs30 < coeffs.vlin) {
      pgaRock = exp(calcMean(
          coeffsPGA, coeffAkBias.correctionPGA, slab(), basin(), adjusted(), akBiasAdjusted(),
          0.0, in.Mw, in.rRup, in.zTor, VS30_ROCK, in.z2p5));
    }
    double μ = calcMean(
        coeffs, coeffAkBias.correction, slab(), basin(), adjusted(), akBiasAdjusted(),
        pgaRock, in.Mw, in.rRup, in.zTor, in.vs30, in.z2p5);

    double σ = calcSigma(coeffs, coeffsPGA, pgaRock, in.rRup, in.vs30);

    double[] μs = GroundMotions.create5th50th95th(μ, coeffs.ε);

    return GroundMotions.createTree(
        GroundMotions.EPI_IDS, μs,
        GroundMotions.EPI_WTS, σ);
  }

  private static double calcMean(
      Coefficients c,
      double correctionAk,
      boolean slab,
      boolean basin,
      boolean adjust,
      boolean akBiasAdjusted,
      double pgaRock,
      double Mw,
      double rRup,
      double zTor,
      double vs30,
      double z2p5) {

    /* Eq. 1,2 - Finite fault term */
    double lnRrupHff = log(rRup + C4 * exp((Mw - 6.0) * c.a9));

    /* Eq. 3,4 - Magnitude scaling */
    double C1 = (slab ? c.c1s : c.c1i);
    double c13m = c.a13 * (10 - Mw) * (10 - Mw);
    double c4s = c.a4 + (slab ? c.a45 : 0.0);
    double fMag = (Mw <= C1 ? c4s : c.a5) * (Mw - C1) + c13m;

    /* Eq. 5 - Additional slab scaling */
    double fSlab = 0.0;
    if (slab) {
      fSlab = c.a10 + (c.a4 + c.a45) * (c.c1s - C1S) + c.a14 * lnRrupHff;
    }

    /* Eq. 6 - Depth scaling: slab only */
    double fDepth = 0.0;
    if (slab) {
      double ΔzTor = (zTor <= 200.0) ? (zTor - 50.0) : 150.0;
      fDepth += ((zTor <= 50.0) ? c.a8 : c.a11) * ΔzTor;
    }

    /* Eq. 7,8 - Site-response scaling */
    double vsS = calcVsStar(vs30);
    double fSite = c.a12 * log(vsS / c.vlin);
    if (vsS < c.vlin) {
      fSite += -c.b * log(pgaRock + C) + c.b * log(pgaRock + C * pow((vsS / c.vlin), N));
    } else {
      fSite += c.b * N * log(vsS / c.vlin);
    }

    /* Eq. 9 - Basin-depth scaling */
    double fBasin = 0.0;
    if (!Double.isNaN(z2p5)) {
      double z2p5m = z2p5 * 1000.0;
      double lnZ2p5ref = 8.52;
      if (vs30 >= 570) {
        lnZ2p5ref = 7.6;
      } else if (vs30 > 200) {
        lnZ2p5ref = 8.52 - 0.88 * log(vs30 / 200.0);
      }

      /* Eq. 11 - Cascadia basin scaling */
      double lnzPrime = calcLnzPrime(z2p5m, exp(lnZ2p5ref));

      if (lnzPrime > 0) {
        if (basin) {
          // USGS T & shallow Z scaling; no need below b/c z2p5>6
          lnzPrime *= GmmUtils.deltaZ25scale(c.imt, z2p5);
        }

        /* Eq. 13 */
        fBasin = c.a39 * lnzPrime;
      }
    }

    /* Eq. 3.1 */
    double μ = c.a1 +
        (c.a2 + c.a3 * (Mw - 7.0)) * lnRrupHff +
        c.a6 * rRup + fMag + fDepth +
        fSite + fSlab +
        fBasin; // + fAs * c.a15; no aftershock term

    if (adjust) {
      μ += c.a1Adj;
    }

    if (!slab && akBiasAdjusted) {
      μ += correctionAk;
    }

    return μ;
  }

  /* Args in meters. */
  private static double calcLnzPrime(double z2p5, double z2p5ref) {
    return log((z2p5 + 50.0) / (z2p5ref + 50.0));
  }

  private static double calcVsStar(double vs30) {
    return min(vs30, VSS_MAX);
  }

  private static double calcSigma(
      Coefficients c,
      Coefficients cPga,
      double pgaRock,
      double rRup,
      double vs30) {

    double τlin = D0;
    double τSq = τlin * τlin;

    double φSq = φlinSq(c.d1, c.d2, rRup);

    double vsS = calcVsStar(vs30);

    /* Add nonlinear effects. */
    if (vsS < c.vlin) {

      /* φamp^2 = 0.3 * 0.3 = 0.09 */
      double φbSq = φSq - 0.09;

      double φlinSqPga = φlinSq(cPga.d1, cPga.d2, rRup);
      double φbSqPga = φlinSqPga - 0.09;

      double dSite = c.b * pgaRock *
          ((1.0 / (pgaRock + C * pow((vs30 / c.vlin), N))) - (1.0 / (pgaRock + C)));
      double dSiteSq = dSite * dSite;

      φSq += dSiteSq * φbSqPga + 2 * dSite * sqrt(φbSqPga) * sqrt(φbSq) * c.ρW;

      /* τlinPga = τlin = D0 */
      τSq += dSiteSq * τSq + 2 * dSite * τSq * c.ρB;
    }

    return sqrt(τSq + φSq);
  }

  private static double φlinSq(double d1, double d2, double rRup) {
    /* Eq. 5.2 */
    double φ = d1;
    if (rRup > 450.0) {
      φ += d2;
    } else if (rRup >= 150.0) {
      φ += d2 * (rRup - 150.0) / 300.0;
    }
    return φ;

  }

  private static abstract class Global extends AbrahamsonGulerce_2020 {
    static final SubductionZone ZONE = GLOBAL;
    static final String NAME = AbrahamsonGulerce_2020.NAME + " : " + ZONE;

    Global(Imt imt, Gmm subtype) {
      super(imt, subtype, ZONE);
    }
  }

  private static abstract class Cascadia extends AbrahamsonGulerce_2020 {
    static final SubductionZone ZONE = CASCADIA;
    static final String NAME = AbrahamsonGulerce_2020.NAME + " : " + ZONE;

    Cascadia(Imt imt, Gmm subtype) {
      super(imt, subtype, ZONE);
    }
  }

  private static abstract class Alaska extends AbrahamsonGulerce_2020 {
    static final SubductionZone ZONE = ALASKA;
    static final String NAME = AbrahamsonGulerce_2020.NAME + " : " + ZONE;

    Alaska(Imt imt, Gmm subtype) {
      super(imt, subtype, ZONE);
    }
  }

  static class GlobalInterface extends Global {
    static final String NAME = Global.NAME + " : " + INTERFACE;

    GlobalInterface(Imt imt) {
      super(imt, AG_20_GLOBAL_INTERFACE);
    }
  }

  static final class GlobalInterfaceAkAdjusted extends GlobalInterface {
    static final String NAME = GlobalInterface.NAME + " (Alaska Adjusted)";

    GlobalInterfaceAkAdjusted(Imt imt) {
      super(imt);
    }

    @Override
    boolean akBiasAdjusted() {
      return true;
    }
  }

  static final class GlobalSlab extends Global {
    static final String NAME = Global.NAME + " : " + SLAB;

    GlobalSlab(Imt imt) {
      super(imt, AG_20_GLOBAL_SLAB);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static class CascadiaInterface extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + INTERFACE;

    CascadiaInterface(Imt imt) {
      this(imt, AG_20_CASCADIA_INTERFACE);
    }

    CascadiaInterface(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }
  }

  static class CascadiaSlab extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + SLAB;

    CascadiaSlab(Imt imt) {
      this(imt, AG_20_CASCADIA_SLAB);
    }

    CascadiaSlab(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static final class CascadiaInterfaceBasin extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " : Basin";

    CascadiaInterfaceBasin(Imt imt) {
      super(imt, AG_20_CASCADIA_INTERFACE_BASIN);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaSlabBasin extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " : Basin";

    CascadiaSlabBasin(Imt imt) {
      super(imt, AG_20_CASCADIA_SLAB_BASIN);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static class CascadiaInterfaceAdjusted extends Cascadia {
    static final String NAME = Cascadia.NAME + " (adj) : " + INTERFACE;

    CascadiaInterfaceAdjusted(Imt imt) {
      this(imt, AG_20_CASCADIA_INTERFACE_ADJUSTED);
    }

    CascadiaInterfaceAdjusted(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }

    @Override
    boolean adjusted() {
      return true;
    }
  }

  static class CascadiaSlabAdjusted extends Cascadia {
    static final String NAME = Cascadia.NAME + " (adj) : " + SLAB;

    CascadiaSlabAdjusted(Imt imt) {
      this(imt, AG_20_CASCADIA_SLAB_ADJUSTED);
    }

    CascadiaSlabAdjusted(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }

    @Override
    boolean slab() {
      return true;
    }

    @Override
    boolean adjusted() {
      return true;
    }
  }

  static final class CascadiaInterfaceAdjustedBasin extends CascadiaInterfaceAdjusted {
    static final String NAME = CascadiaInterfaceAdjusted.NAME + " : Basin";

    CascadiaInterfaceAdjustedBasin(Imt imt) {
      super(imt, AG_20_CASCADIA_INTERFACE_ADJUSTED_BASIN);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaSlabAdjustedBasin extends CascadiaSlabAdjusted {
    static final String NAME = CascadiaSlabAdjusted.NAME + " : Basin";

    CascadiaSlabAdjustedBasin(Imt imt) {
      super(imt, AG_20_CASCADIA_SLAB_ADJUSTED_BASIN);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class AlaskaInterface extends Alaska {
    static final String NAME = Alaska.NAME + " : " + INTERFACE;

    AlaskaInterface(Imt imt) {
      super(imt, AG_20_ALASKA_INTERFACE);
    }
  }

  static final class AlaskaSlab extends Alaska {
    static final String NAME = Alaska.NAME + " : " + SLAB;

    AlaskaSlab(Imt imt) {
      super(imt, AG_20_ALASKA_SLAB);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static final class AlaskaInterfaceAdjusted extends Alaska {
    static final String NAME = Alaska.NAME + " (adj) : " + INTERFACE;

    AlaskaInterfaceAdjusted(Imt imt) {
      super(imt, AG_20_ALASKA_INTERFACE_ADJUSTED);
    }

    @Override
    boolean adjusted() {
      return true;
    }
  }

  static final class AlaskaSlabAdjusted extends Alaska {
    static final String NAME = Alaska.NAME + " (adj) : " + SLAB;

    AlaskaSlabAdjusted(Imt imt) {
      super(imt, AG_20_ALASKA_SLAB_ADJUSTED);
    }

    @Override
    boolean slab() {
      return true;
    }

    @Override
    boolean adjusted() {
      return true;
    }
  }
}
