package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.ALASKA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.CASCADIA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.GLOBAL;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.SLAB;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.max;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable.Position;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the Kuehn et al. (2020) ground motion model for subduction
 * regions developed as part of the <a
 * href="https://www.risksciences.ucla.edu/nhr3/nga-subduction">NGA
 * Subduction</a> project.
 *
 * <p><b>Implementation notes:</b>
 *
 * <ul><li>Implementations are provided for the global model as well as Cascadia
 * and Alaska regionalized models. Implementations include USGS developed depth
 * tapering and restriction to long periods ({@code T &geq; 1 s}), as well as a
 * Seattle-specific basin implementation that makes further adjustments based on
 * M9 simulations for interface events at sites over the deepest depths
 * ({@code Z2.5 > 6 km}) and at long periods ({@code T &qeq; 2 s}). Specialized
 * flags are now supported as GMM inputs and these may also trigger
 * Seattle-specific behavior in the Cascadia interface implementation.
 *
 * <li>Simplified epistemic uncertainty tables for the model were obtained from
 * <a href="https://github.com/nikuehn/KBCG20/tree/master/UNCERTAINTY">
 * https://github.com/nikuehn/KBCG20/tree/master/UNCERTAINTY</a>.
 *
 * <li>The model, as published does not include a basin term for Alaska, however
 * if {@code z}-values are supplied then the general Cascadia basin term is
 * used.
 *
 * <li>For the 2023 NSHM, implementations were added that support Seattle M9
 * interface simulation informed modifications to basin amplification terms.
 * This is a region-specific behavior and is managed through the use of flags
 * that are passed in with {@code GmmInput}. Note that the Cascadia model
 * includes specialized behavior (no M9) by default for both slab and
 * interface.</ul>
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Reference:</b> Kuehn, N., Bozorgnia, Y, Campbell, K.W., Gregor, N.
 * (2020) Partially non-ergodic ground-motion model for subduction regions using
 * the NGA-Subduction database. Report no. 2020/04. Berkeley, CA: PEER, 147 p.
 *
 * <p><b>Component:</b> average horizontal (RotD50)
 *
 * @author U.S. Geological Survey
 */

public abstract class KuehnEtAl_2020 implements GroundMotionModel {

  /*
   *
   * Developer notes:
   *
   * (Some of the issues noted below will likely be cleared up when an
   * Earthquake Spectra paper is published, as was the case with Parker et al.)
   *
   * δMbreak, defined in Figure 4.2. The paper does not mention regionalization
   * for this parameter, but in the Python code only Japan and South America
   * have non-zero values.
   *
   * Source Zref(if|slab) are not specified in the paper. Python code specifies
   * 15 and 50 km, respectively
   *
   * Eq. 4.2 - median PSA at short periods is not allowed to be less than PGA,
   * so, if PSA(T&lt;0.1) &lt; PGA, do we still use sigma(PSA(T&lt;0.1)) or use
   * sigma(PGA)?
   *
   * Eq. 4.8 in the paper includes `c ln( vs30/k1)^n`, this ln is omitted in the
   * Python and the other papers using this term (CB14, etc.)
   *
   * I think fSite term in Python code has a typo, using `vs30 &lt; k1` rather
   * than `vs30 ≤ k1` as specified in the paper (and others using the same site
   * term)
   *
   * Value of k2 (hard-coded in Python) for PGV is a typo, the value should be
   * -1.955 but -1.995 is used
   *
   * basin() indicates USGS T and z limitations
   *
   * seattle() indicates site inside Seattle polygon and deep M9 scaling
   *
   * Kuehn Cascadia models automatically include Kuehn Seattle behavior (is this
   * accurate?)
   *
   * Coefficients for Alaska (theta_1_slab_reg_Al, theta_1_if_reg_Al,
   * theta_7_reg_Al, theta_6_2_reg_Al) and mean_redisual_seattle_basin updated
   * based on Kuehn's GitHub repository:
   *
   * https://github.com/nikuehn/KBCG20/blob/master/COEFFS_SEP21/
   * coefficients_KBCG20.csv
   *
   * 11/06/2023 Added non-linear site adjustment to sigma per KBCG authors at
   * Alaska NSHM tiger team meeting. This has not been approved for use in CONUS
   * model, only Alaska
   */

  static final String NAME = "Kuehn et al. (2020)";

  /*
   * Note that author-defined constraints for magnitude breakpoint, source
   * depth, forearc/backarc subregions, basin depth are region-specific.
   */
  static final Constraints CONSTRAINTS_INTERFACE = Constraints.builder()
      .set(MW, Range.closed(5.0, 9.5))
      .set(RRUP, Range.closed(10.0, 1000.0))
      .set(ZTOR, Range.closed(0.0, 50.0))
      .set(VS30, Range.closed(150.0, 1500.0))
      .build();

  static final Constraints CONSTRAINTS_SLAB = Constraints.builder()
      .set(MW, Range.closed(5.0, 8.5))
      .set(RRUP, Range.closed(10.0, 1000.0))
      .set(ZTOR, Range.closed(0.0, 200.0)) // for Columbia: ZTOR <= 150
      .set(VS30, Range.closed(150.0, 1500.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("KBCG20.csv");
  static final CoefficientContainer COEFFS_AK_BIAS =
      new CoefficientContainer("nga-sub-ak-interface-bias-corr.csv");

  private static final double δM = 0.1; // p. 20
  private static final double MREF = 6.0; // p. 20
  private static final double δZ = 1.0; // p. 22
  private static final double C = 1.88; // from CB14
  private static final double N = 1.18; // from CB14
  private static final double VSROCK = 1100;
  private static final double PHI_LNAF_SQ = 0.09; // 0.3^2

  private static final class Coefficients {
    final Imt imt;
    final SubductionZone zone;

    final double θ1;
    final double θ2;
    final double θ3;
    final double θ4;
    final double θ5;
    final double θ6_2;
    final double θ7;
    final double θ9;
    final double θ10;
    final double θ11;
    final double θ11S;
    final double θ12;

    final double θnft1;
    final double θnft2;

    final double k1;
    final double k2;

    final double φ;
    final double τ;
    final double ρ;

    final double mBreak;
    final double zBreak; // dzb_if dzb_slab
    final double zRef;

    Coefficients(Imt imt, CoefficientContainer cc, SubductionZone zone, boolean slab) {
      this.imt = imt;
      this.zone = zone;
      Map<String, Double> coeffs = cc.get(imt);

      θ3 = coeffs.get("theta_3");
      θ5 = coeffs.get("theta_5");

      if (slab) {
        θ2 = coeffs.get("theta_2_slab");
        θ4 = coeffs.get("theta_4_slab");
        θ9 = coeffs.get("theta_9_slab");

        zBreak = 80.0 + coeffs.get("dzb_slab");
        zRef = 50.0; // this value is not in the paper
      } else {
        θ2 = coeffs.get("theta_2_if");
        θ4 = coeffs.get("theta_4_if");
        θ9 = coeffs.get("theta_9_if");

        zBreak = 30.0 + coeffs.get("dzb_if");
        zRef = 15.0; // this value is not in the paper
      }

      /* For now we apply general Cascadia basin scaling elsewhere. */
      θ11 = coeffs.get("intercept_Ca");
      θ11S = coeffs.get("mean_residual_Seattle_basin");
      θ12 = coeffs.get("slope_Ca");

      if (zone == CASCADIA) {
        θ1 = (slab) ? coeffs.get("theta_1_slab_reg_Ca") : coeffs.get("theta_1_if_reg_Ca");
        θ6_2 = coeffs.get("theta_6_2_reg_Ca");
        θ7 = coeffs.get("theta_7_reg_Ca");
      } else if (zone == ALASKA) {
        θ1 = (slab) ? coeffs.get("theta_1_slab_reg_Al") : coeffs.get("theta_1_if_reg_Al");
        θ6_2 = coeffs.get("theta_6_2_reg_Al");
        θ7 = coeffs.get("theta_7_reg_Al");
      } else { // global
        θ1 = (slab) ? coeffs.get("mu_theta_1_slab") : coeffs.get("mu_theta_1_if");
        θ6_2 = coeffs.get("mu_theta_6");
        θ7 = coeffs.get("mu_theta_7");
      }

      θ10 = 0.0; // Table 4.1: fixed to 0
      θnft1 = coeffs.get("nft_1");
      θnft2 = coeffs.get("nft_2");

      k1 = coeffs.get("k1");
      k2 = coeffs.get("k2");

      double mb = slab ? 7.6 : 7.9; // global
      if (zone == CASCADIA)
        mb = slab ? 7.2 : 8.0;
      else if (zone == ALASKA) {
        mb = slab ? 7.2 : 8.6;
      }
      mBreak = mb;

      φ = coeffs.get("phi");
      τ = coeffs.get("tau");
      ρ = coeffs.get("rho");
    }
  }

  /*
   * Average bias-correction for NGA-Subduction interface models in Alaska nshm.
   */
  private static final class CoefficientsAkBias {
    final double correction;
    final double correctionPGA;

    CoefficientsAkBias(Imt imt, CoefficientContainer cc) {
      correction = cc.get(imt).get("bias_ak");
      correctionPGA = cc.get(PGA).get("bias_ak");
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;
  private final CoefficientsAkBias coeffsAkBias;
  private final GroundMotionTable epistemicTable;

  KuehnEtAl_2020(Imt imt, SubductionZone zone) {
    coeffs = new Coefficients(imt, COEFFS, zone, slab());
    coeffsPGA = new Coefficients(PGA, COEFFS, zone, slab());
    coeffsAkBias = new CoefficientsAkBias(imt, COEFFS_AK_BIAS);
    SubductionZone.Type szType = slab() ? SLAB : INTERFACE;
    epistemicTable = GroundMotionTables.getKbcg20(zone, szType, imt);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  /* Intraslab (true) or interface (false) model. */
  boolean slab() {
    return false;
  }

  /* Use USGS deep basin, long period amplification model. */
  boolean basin() {
    return false;
  }

  /* Apply Keuhn et al. Seattle basin. */
  boolean seattle() {
    return false;
  }

  /* Apply Seattle M9 adjustment. */
  boolean m9() {
    return false;
  }

  /* Apply USGS bias adjustment for Alaska */
  boolean akBiasAdjusted() {
    return false;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {

    double pgaRef = exp(calcPgaRef(
        coeffsPGA, coeffsAkBias.correctionPGA, slab(), akBiasAdjusted(),
        in.Mw, in.rRup, in.zTor));

    double μ = calcMean(
        coeffs, coeffsAkBias.correction, slab(), basin(), seattle(), m9(), akBiasAdjusted(),
        in.Mw, in.rRup, in.zTor, in.vs30, in.z2p5, pgaRef);

    // short periods can't be lower than PGA
    if (coeffs.imt.isSA() && coeffs.imt.period() <= 0.1) {
      double μPga = calcMean(
          coeffsPGA, coeffsAkBias.correctionPGA, slab(), basin(), seattle(), m9(), akBiasAdjusted(),
          in.Mw, in.rRup, in.zTor, in.vs30, in.z2p5, pgaRef);
      μ = max(μ, μPga);
    }

    // Tiger team recommendation to use CB14 sigma that
    // accounts for non-linear amplification
    double σ = calcStdDev(coeffs, coeffsPGA, in.vs30, pgaRef);

    Position p = epistemicTable.position(in.rRup, in.Mw);
    double ψ = epistemicTable.get(p);

    double[] μs = GroundMotions.create5th50th95th(μ, ψ);

    return GroundMotions.createTree(
        GroundMotions.EPI_IDS, μs,
        GroundMotions.EPI_WTS, σ);
  }

  private static double calcPgaRef(
      Coefficients c,
      double correctionAk,
      boolean slab,
      boolean akBiasAdjusted,
      double Mw,
      double rRup,
      double zTor) {

    double fMag = magnitudeTerm(c, Mw);
    double fGeom = geometricAttenuationTerm(c, Mw, rRup);
    double fDepth = depthTerm(c, zTor);
    double fAtten = arcCrossingTerm(c, rRup);
    double fSite = siteAmpTerm(c, VSROCK, 0);
    double μ = c.θ1 + fMag + fGeom + fDepth + fAtten + fSite;

    if (!slab && akBiasAdjusted) {
      μ += correctionAk;
    }

    return μ;
  }

  private static double calcMean(
      Coefficients c,
      double correctionAk,
      boolean slab,
      boolean basin,
      boolean seattle,
      boolean m9,
      boolean akBiasAdjusted,
      double Mw,
      double rRup,
      double zTor,
      double vs30,
      double z2p5,
      double pgaRef) {

    double fMag = magnitudeTerm(c, Mw);
    double fGeom = geometricAttenuationTerm(c, Mw, rRup);
    double fDepth = depthTerm(c, zTor);
    double fAtten = arcCrossingTerm(c, rRup);
    double fSite = siteAmpTerm(c, vs30, pgaRef);
    double fBasin = basinTerm(c, vs30, z2p5, slab, basin, seattle, m9);
    double μ = c.θ1 + fMag + fGeom + fDepth + fAtten + fSite + fBasin;

    if (!slab && akBiasAdjusted) {
      μ += correctionAk;
    }

    return μ;
  }

  /* Eq. 4.3 used for Mw and zTor */
  private static double loghinge(
      double x,
      double x0,
      double a,
      double b0,
      double b1,
      double δ) {

    return a + b0 * (x - x0) + (b1 - b0) * δ * log(1 + exp((x - x0) / δ));
  }

  /* Eq. 4.4 - magnitude term */
  private static double magnitudeTerm(
      Coefficients c,
      double Mw) {

    return loghinge(Mw, c.mBreak, c.θ4 * (c.mBreak - MREF), c.θ4, c.θ5, δM);
  }

  /*
   * Eq. 4.5 - geometrical attenuation term
   *
   * Eq. 4.6 - finite fault term ("pow(10,...)")
   */
  private static double geometricAttenuationTerm(
      Coefficients c,
      double Mw,
      double rRup) {

    double ffExp = c.θnft1 + c.θnft2 * (Mw - 6.0);
    return (c.θ2 + c.θ3 * Mw) * log(rRup + pow(10, ffExp));
  }

  /* Eq. 4.7 - source depth term */
  private static double depthTerm(
      Coefficients c,
      double zTor) {

    return loghinge(zTor, c.zBreak, c.θ9 * (c.zBreak - c.zRef), c.θ9, c.θ10, δZ);
  }

  /* Eq. 4.8 - site amplification term */
  private static double siteAmpTerm(Coefficients c, double vs30, double pgaRef) {
    double siteTerm;
    if (vs30 <= c.k1) {
      siteTerm = c.θ7 * log(vs30 / c.k1) +
          c.k2 * (log(pgaRef + C * pow(vs30 / c.k1, N)) - log(pgaRef + C));
    } else {
      siteTerm = (c.θ7 + c.k2 * N) * log(vs30 / c.k1);
    }
    return siteTerm;
  }

  /*
   * Eq. 4.9 - anelastic attenuation term (arc-crossing term) -- Region specific
   * (regionalization ignored for now). This is accurate/applicable for ALASKA,
   * CASCADIA, NZ, and TW.
   */
  private static double arcCrossingTerm(Coefficients c, double rRup) {
    return c.θ6_2 * rRup;
  }

  /* Eq. 4.10 - z2.5 basin depth term; Cascadia model */
  private static double basinTerm(
      Coefficients c,
      double vs30,
      double z2p5,
      boolean slab,
      boolean basin,
      boolean seattle,
      boolean m9) {

    if (Double.isNaN(z2p5)) {
      return 0.0;
    }

    double z2p5m = z2p5 * 1000.0;
    double δlnZ = log(z2p5m) - basinZref(vs30);

    double fb = 0.0;

    /*
     * Currently there is not a smooth transition from default basin depth
     * scaling to Kuehn's fixed amplification factor for the Seattle basin nor
     * from the Seattle basin to the M9 ln(2) scale factor, which is larger than
     * most c.θ11S at T≥2 s.
     */
    if (seattle) { // basin is implicitely true for seattle() and m9()
      fb = c.θ11S; // period-dependent seattle scale factor
      if (m9 && !slab && z2p5 > 6.0 && (c.imt.isSA() && c.imt.period() > 1.9)) {
        fb = log(2.0); // M9 scaling
      }
    } else {
      fb = Math.min(c.θ11 + c.θ12 * δlnZ, c.θ11S); // Seattle cap
      /*
       * NOTE: If we stop using the USGS basin model in NSHMs, then at shorter
       * periods c.θ11S can often impose significant deamplification, more than
       * c.θ11 + c.θ12 * δlnZ and this was not likely the intent of the
       * developers.
       */
    }
    if (basin) {
      fb *= GmmUtils.deltaZ25scale(c.imt, z2p5);
    }
    return fb;
  }

  /* Eq. 4.11 - basin reference depth for vs30; Cascadia model */
  private static double basinZref(double vs30) {
    double θz1 = 8.29404964010203;
    double θz2 = 2.30258509299405;
    double θz3 = 6.39692965521615;
    double θz4 = 0.27081459;
    double vsterm = exp((log(vs30) - θz3) / θz4);
    return θz1 + (θz2 - θz1) * vsterm / (1.0 + vsterm);
  }

  /*
   * Aleatory uncertainty model with adjustments to sigma after CB14 for
   * non-linear site amplification.
   */
  private static double calcStdDev(
      Coefficients c,
      Coefficients cPGA,
      double vs30,
      double pgaRock) {

    // -- Equation 31
    double vsk1 = vs30 / c.k1;
    double alpha = (vs30 < c.k1) ? c.k2 * pgaRock *
        (1 / (pgaRock + C * pow(vsk1, N)) - 1 / (pgaRock + C)) : 0.0;

    // no small magnitude dependence (eq 27 & 28 CB14)
    double tau_lnYB = c.τ;
    double phi_lnY = c.φ;
    double tau_lnPGAB = cPGA.τ;
    double phi_lnPGAB = cPGA.φ;

    // inter-event std dev -- Equation 29
    double alphaTau = alpha * tau_lnPGAB;
    double tauSq = tau_lnYB * tau_lnYB + alphaTau * alphaTau +
        2.0 * alpha * c.ρ * tau_lnYB * tau_lnPGAB;

    // intra-event std dev -- Equation 30
    double phi_lnYB = sqrt(phi_lnY * phi_lnY - PHI_LNAF_SQ);
    phi_lnPGAB = sqrt(phi_lnPGAB * phi_lnPGAB - PHI_LNAF_SQ);
    double aPhi_lnPGAB = alpha * phi_lnPGAB;

    // phi_lnaf terms in eqn. 30 cancel when expanded leaving phi_lnY only
    double phiSq = phi_lnY * phi_lnY + aPhi_lnPGAB * aPhi_lnPGAB +
        2.0 * c.ρ * phi_lnYB * aPhi_lnPGAB;

    // total model -- Equation 32
    return sqrt(phiSq + tauSq);
  }

  private static abstract class Global extends KuehnEtAl_2020 {
    static final SubductionZone ZONE = GLOBAL;
    static final String NAME = KuehnEtAl_2020.NAME + " : " + ZONE;

    Global(Imt imt) {
      super(imt, ZONE);
    }
  }

  private static abstract class Cascadia extends KuehnEtAl_2020 {
    static final SubductionZone ZONE = CASCADIA;
    static final String NAME = KuehnEtAl_2020.NAME + " : " + ZONE;

    Cascadia(Imt imt) {
      super(imt, ZONE);
    }
  }

  private static abstract class Alaska extends KuehnEtAl_2020 {
    static final SubductionZone ZONE = ALASKA;
    static final String NAME = KuehnEtAl_2020.NAME + " : " + ZONE;

    Alaska(Imt imt) {
      super(imt, ZONE);
    }
  }

  static final class GlobalInterface extends Global {
    static final String NAME = Global.NAME + " : " + INTERFACE;

    GlobalInterface(Imt imt) {
      super(imt);
    }
  }

  static final class GlobalInterfaceAkAdjusted extends Global {
    static final String NAME = GlobalInterface.NAME + " (Alaska Adjusted)";

    GlobalInterfaceAkAdjusted(Imt imt) {
      super(imt);
    }

    @Override
    boolean akBiasAdjusted() {
      return true;
    }
  }

  static class GlobalSlab extends Global {
    static final String NAME = Global.NAME + " : " + SLAB;

    GlobalSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  /*
   * 2023-02-23 @pmpowers: Keuhn's Seattle basin model and the M9 adjustments
   * are both dependent on a site being inside the Seattle basin polygon. The M9
   * adjustment further depends on T≥2 s and z2.5>6 km. Kuehn's Seattle basin
   * model applies everywhere within the Seattle polygon and is subject to USGS
   * period and depth scaling.
   *
   * TODO need figure of what the depth tapering looks like for Kuehn's model.
   *
   * CascadiaInterface ------------> (no flags)
   *
   * CascadiaInterfaceBasin -------> basin() (USGS basin model)
   *
   * CascadiaInterfaceSeattle -----> basin() seattle() (Seattle basin)
   *
   * CascadiaInterfaceSeattleM9 ---> basin() seattle() m9() (Seattle basin + M9)
   *
   *
   * CascadiaSlab -----------------> (no flags)
   *
   * CascadiaSlabBasin ------------> basin()
   *
   * CascadiaSlabSeattle ----------> seattle() (Seattle basin)
   */
  static class CascadiaInterface extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + INTERFACE;

    CascadiaInterface(Imt imt) {
      super(imt);
    }
  }

  static class CascadiaInterfaceBasin extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " : Basin";

    CascadiaInterfaceBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static class CascadiaInterfaceSeattleBasin extends CascadiaInterfaceBasin {
    static final String NAME = CascadiaInterface.NAME + " : Seattle Basin";

    CascadiaInterfaceSeattleBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean seattle() {
      return true;
    }
  }

  static final class CascadiaInterfaceSeattleBasinM9 extends CascadiaInterfaceSeattleBasin {
    static final String NAME = CascadiaInterfaceSeattleBasin.NAME + " : M9";

    CascadiaInterfaceSeattleBasinM9(Imt imt) {
      super(imt);
    }

    @Override
    boolean m9() {
      return true;
    }

  }

  static class CascadiaSlab extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + SLAB;

    CascadiaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static final class CascadiaSlabBasin extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " : Basin";

    CascadiaSlabBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaSlabSeattleBasin extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " : Seattle Basin";

    CascadiaSlabSeattleBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean seattle() {
      return true;
    }
  }

  static final class AlaskaInterface extends Alaska {
    static final String NAME = Alaska.NAME + " : " + INTERFACE;

    AlaskaInterface(Imt imt) {
      super(imt);
    }
  }

  static final class AlaskaSlab extends Alaska {
    static final String NAME = Alaska.NAME + " : " + SLAB;

    AlaskaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

}
