package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.ALASKA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.CASCADIA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.GLOBAL;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.SLAB;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.log10;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the Parker et al. (2022) ground motion model for subduction
 * regions developed as part of the <a
 * href="https://www.risksciences.ucla.edu/nhr3/nga-subduction">NGA
 * Subduction</a> project.
 *
 * <p><b>Implementation notes:</b>
 *
 * <ul><li>Model currently uses {@code zTor} for intraslab depth scaling but
 * authors recommend {@code zHyp}.
 *
 * <li>The basin depth scaling model is that developed for Cascadia.
 * Implementations include USGS developed depth tapering and restriction to long
 * periods ({@code T &geq; 1 s}), as well as a Seattle-specific basin
 * implementation that makes further adjustments based on M9 simulations for
 * interface events at sites over the deepest depths ({@code Z2.5 > 6 km}) and
 * at long periods ({@code T &qeq; 2 s}). Specialized flags are now supported as
 * GMM inputs and these may also trigger Seattle-specific behavior in the
 * Cascadia interface implementation.
 *
 * <li>Implementations are provided for the global model as well as Cascadia and
 * Alaska regionalized models.
 *
 * <li>For the 2023 NSHM, implementations were added that support Seattle M9
 * interface simulation informed modifications to basin amplification terms.
 * This is a region-specific behavior and is managed through the use of flags
 * that are passed in with {@code GmmInput}.</ul>
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Reference:</b> Parker, G.A, Stewart, J.P., Boore, D.M., Atkinson, G.M.,
 * and Hassani, B. (2022) NGA-Subduction global ground motion models with
 * regional adjustment factors, Earthquake Spectra, v. 38, n. 1, p. 456-493,
 * doi: <a
 * href="https://doi.org/10.1177/87552930211034889">10.1177/87552930211034889</a>.
 *
 * <p><b>Reference:</b> Parker, G.A, Stewart, J.P., Boore, D.M., Atkinson, G.M.,
 * and Hassani, B. (2020) NGA-Subduction global ground motion models with
 * regional adjustment factors. Report no. 2020/03. Berkeley, CA: PEER, 131 p.
 *
 * <p><b>Component:</b> average horizontal (RotD50)
 *
 * @author U.S. Geological Survey
 */

public abstract class ParkerEtAl_2020 implements GroundMotionModel {

  /*
   * Developer notes:
   *
   * Regions: global, Alaska, Cascadia, [CAM (Central America), Japan, SA (South
   * America), Taiwan]
   *
   * Saturation Regions: global, Aleutian, Alaska, Cascadia, [Central America S,
   * Central America N, Japan Pacific, Japan Phillipines, South America N, South
   * America S, Taiwan W, Taiwan E]
   *
   * Implementation currently uses zTor for intraslab depth scaling but authors
   * regressed on zHyp. Which parameter is used will possibly depend on
   * implementation choices for the 2023 NSHM. Historically, intraslab GMMs have
   * used zTor (check this) and zHyp makes depths a little deeper, assuming
   * pseudo faults have a measureable width, and would yield slightly higher
   * ground motions).
   */

  static final String NAME = "Parker et al. (2020)";

  static final Constraints CONSTRAINTS_INTERFACE = Constraints.builder()
      .set(MW, Range.closed(4.5, 9.5))
      .set(RRUP, Range.closed(20.0, 1000.0))
      .set(ZTOR, Range.closed(0.0, 40.0)) // should be ZHYP
      .set(VS30, Range.closed(150.0, 2000.0))
      .build();

  static final Constraints CONSTRAINTS_SLAB = Constraints.builder()
      .set(MW, Range.closed(4.5, 8.5))
      .set(RRUP, Range.closed(35.0, 1000.0))
      .set(ZTOR, Range.closed(0.0, 200.0))
      .set(VS30, Range.closed(150.0, 2000.0))
      .build();

  static final CoefficientContainer COEFFS_INTERFACE;
  static final CoefficientContainer COEFFS_SLAB;
  static final CoefficientContainer COEFFS_AK_BIAS;

  static {
    /*
     * Aleatory uncertainty coefficients are duplicated in interface and
     * interface CSV files (as are some other coefficients that are also
     * independent of event type)
     */
    COEFFS_INTERFACE = new CoefficientContainer("PSBAH20_interface.csv");
    COEFFS_SLAB = new CoefficientContainer("PSBAH20_slab.csv");
    COEFFS_AK_BIAS = new CoefficientContainer("nga-sub-ak-interface-bias-corr.csv");
  }

  private static final double B4 = 0.1;
  private static final double V1 = 270.0; // m/s
  private static final double VREF = 760.0; // m/s

  private static final double LN_Z2P5_REF_PARKER = log(1279.0);
  private static final double Z_SLAB = 6.5; // km
  private static final double θ_W = 0.48; // eq. 15

  private static final double DB1 = 20.0; // km
  private static final double DB2 = 67.0; // km

  /* ϕ parameters */
  private static final double ΦR1 = 200; // km
  private static final double ΦR2 = 500; // km
  private static final double ΦV1 = 200; // m/s
  private static final double ΦV2 = 500; // m/s

  private static final class Coefficients {
    final Imt imt;
    final double c0; // model constant
    final double c1, a0; // path scaling
    final double mc; // saturation corner magnitude
    final double c4, c5, c6; // magnitude scaling
    final double d, m; // source depth scaling (slab only)
    final double v2, s1, s2; // linear site amplification
    final double f4, f5; // non-linear site amplification
    final double c_e1, c_e2, c_e3; // Cascadia basin
    final double τ, ϕ21, ϕ22, ϕ2v;
    final double σε1, σε2, t1, t2;

    // inlined
    // final double b4, v1, vRef;

    // unused
    // final double db;
    // final del_None, del_Seattle;
    // final double vm, ϕ2s2s0, a1, ϕ2ss1, ϕ2ss2, a2;

    Coefficients(Imt imt, CoefficientContainer cc, SubductionZone zone, boolean slab) {

      this.imt = imt; // Required for σε

      Map<String, Double> coeffs = cc.get(imt);

      // Model constant
      c0 = coeffs.get(zone + "_c0");

      // Path
      c1 = coeffs.get("c1");
      // b4 inlined as constant

      // a0 is region-specific for Alaska and, for slab events, Cascadia,
      // otherwise use Global value
      a0 = coeffs.get(zone + "_a0");

      // Magnitude saturation corner magnitude
      // Table 4.1
      if (zone == CASCADIA) {
        mc = slab ? 7.20 : 7.7;
      } else if (zone == ALASKA) {
        mc = slab ? 7.20 : 8.6;
      } else {
        // global
        mc = slab ? 7.6 : 7.9;
      }

      // Magnitude-scaling
      c4 = coeffs.get("c4");
      c5 = coeffs.get("c5");
      c6 = coeffs.get("c6");

      // Source depth scaling (slab only)
      d = slab ? coeffs.get("d") : 0.0;
      m = slab ? coeffs.get("m") : 0.0;

      // Linear site amplification
      // v1 inlined as constant
      v2 = coeffs.get("V2");;
      // vRef inlined as constant

      // s2 is region specific, except for Aleutians (not
      // currently supported), which uses the Global value
      s2 = coeffs.get(zone + "_s2");
      s1 = s2; // s1 = s2 except for Japan and Taiwan

      // Non-linear site amplification
      f4 = coeffs.get("f4");
      f5 = coeffs.get("f5");

      // Cascadia basin
      c_e1 = coeffs.get("C_e1");
      c_e2 = coeffs.get("C_e2");
      c_e3 = coeffs.get("C_e3");

      // aleatory model
      τ = coeffs.get("Tau");
      ϕ21 = coeffs.get("phi21");
      ϕ22 = coeffs.get("phi22");
      ϕ2v = coeffs.get("phi2V");

      // epistemic model, Table E4
      if (zone == CASCADIA) {
        σε1 = slab ? 0.35 : 0.43;
        σε2 = slab ? 0.16 : 0.33;
        t1 = slab ? 0.20 : 0.20;
        t2 = slab ? 3.00 : 0.50;
      } else if (zone == ALASKA) {
        σε1 = slab ? 0.15 : 0.15;
        σε2 = slab ? 0.12 : 0.10;
        t1 = slab ? 0.50 : 1.00;
        t2 = slab ? 1.00 : 4.00;
      } else if (zone == GLOBAL) {
        σε1 = slab ? 0.35 : 0.40;
        σε2 = slab ? 0.22 : 0.40;
        t1 = slab ? 0.15 : 0.20;
        t2 = slab ? 2.00 : 0.40;
      } else {
        throw new IllegalArgumentException("Subduction zone [" + zone + "] not supported");
      }
    }
  }

  /*
   * Average bias-correction for NGA-Subduction interface models in Alaska nshm.
   */
  private static final class CoefficientsAkBias {
    final double correction;
    final double correctionPGA;

    CoefficientsAkBias(Imt imt, CoefficientContainer cc) {
      correction = cc.get(imt).get("bias_ak");
      correctionPGA = cc.get(PGA).get("bias_ak");
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;
  private final CoefficientsAkBias coeffAkBias;

  ParkerEtAl_2020(Imt imt, SubductionZone zone) {
    CoefficientContainer cc = slab() ? COEFFS_SLAB : COEFFS_INTERFACE;
    coeffs = new Coefficients(imt, cc, zone, slab());
    coeffsPGA = new Coefficients(PGA, cc, zone, slab());
    coeffAkBias = new CoefficientsAkBias(imt, COEFFS_AK_BIAS);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  /* Intraslab (true) or interface (false) model. */
  boolean slab() {
    return false;
  }

  /* Use USGS deep basin, long period amplification model. */
  boolean basin() {
    return false;
  }

  /* Apply Seattle M9 adjustment. */
  boolean m9() {
    return false;
  }

  /* Apply USGS bias adjustment for Alaska */
  boolean akBiasAdjusted() {
    return false;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {

    double pgaRef = exp(calcMean(
        coeffsPGA, coeffAkBias.correctionPGA, slab(), akBiasAdjusted(), in));

    double μ = calcMean(
        coeffs, coeffAkBias.correction, slab(), basin(), m9(), akBiasAdjusted(), pgaRef, in);

    double ϕTotal = phiTotal(coeffs, in.rRup, in.vs30);
    double σ = Maths.hypot(coeffs.τ, ϕTotal);
    double ε = calcEpistemic(coeffs);

    double[] μs = GroundMotions.create5th50th95th(μ, ε);

    return GroundMotions.createTree(
        GroundMotions.EPI_IDS, μs,
        GroundMotions.EPI_WTS, σ);
  }

  /* Calc median at 760 reference condition */
  private static double calcMean(
      Coefficients c,
      double correctionAk,
      boolean slab,
      boolean akBiasAdjusted,
      GmmInput in) {

    double h = nearSourceSaturation(in.Mw, c.mc, slab);
    double Fp = pathTerm(c, in.rRup, in.Mw, h);
    double Fm = magnitudeTerm(c, in.Mw);
    double Fd = depthTerm(c, slab, in.zTor);
    double μ = c.c0 + Fp + Fm + Fd;

    if (!slab && akBiasAdjusted) {
      μ += correctionAk;
    }

    return μ;
  }

  /* Equation 1 */
  private static double calcMean(
      Coefficients c,
      double correctionAk,
      boolean slab,
      boolean basin,
      boolean m9,
      boolean akBiasAdjusted,
      double pgaRef,
      GmmInput in) {

    double h = nearSourceSaturation(in.Mw, c.mc, slab);
    double Fp = pathTerm(c, in.rRup, in.Mw, h);
    double Fm = magnitudeTerm(c, in.Mw);
    double Fd = depthTerm(c, slab, in.zTor);

    /* Equation 7 */
    double Fslin = siteTermLinear(c, in.vs30);
    double Fsnl = siteTermNonLinear(c, pgaRef, in.vs30);
    double Fsb = siteBasinTerm(c, in.vs30, in.z2p5, slab, basin, m9);
    double Fs = Fslin + Fsnl + Fsb;

    double μ = c.c0 + Fp + Fm + Fd + Fs;

    if (!slab && akBiasAdjusted) {
      μ += correctionAk;
    }

    return μ;
  }

  /* Equation 2, 3 */
  private static double pathTerm(Coefficients c, double rRup, double Mw, double h) {
    double Rref = Maths.hypot(1, h);
    double R = Maths.hypot(rRup, h);
    double R_Rref = R / Rref;
    return c.c1 * log(R) + B4 * Mw * log(R_Rref) + c.a0 * R;
  }

  private static double nearSourceSaturation(double Mw, double mc, boolean slab) {
    if (slab) {
      /* Equation 4b */
      return (Mw <= mc) ? pow(10, (1.050 / (mc - 4.0)) * (Mw - mc) + 1.544) : 35;
    } else {
      /* Equation 4a */
      return pow(10.0, -0.82 + 0.252 * Mw);
    }
  }

  /* Equation 5 */
  private static double magnitudeTerm(Coefficients c, double Mw) {
    double dM = Mw - c.mc;
    return (Mw <= c.mc) ? c.c4 * dM + c.c5 * dM * dM : c.c6 * dM;
  }

  /* Equation 6 */
  private static double depthTerm(Coefficients c, boolean slab, double zTor) {
    if (!slab) {
      return 0.0;
    }

    /*
     * Most slab GMMs regress on zTor. Currently PointSources have a fixed width
     * of 10 km (not accounting for slab dip) and so we end up with zHyp of zTor
     * + 5 km. This needs work at the PointSource level so we instead follow
     * Parker et al. (2021) guidance on estimating zHyp from zTor. We assume a
     * 6.5 km thick (vertical) slab and a vertical fault and adjust based on
     * Parker's zHyp_bar guidance.
     */
    double zHyp = zTor + θ_W * Z_SLAB;

    double dTerm;
    if (zHyp < DB1) {
      dTerm = c.m * (DB1 - DB2) + c.d;
    } else if (zHyp > DB2) {
      dTerm = c.d;
    } else {
      dTerm = c.m * (zHyp - DB2) + c.d;
    }
    return dTerm;
  }

  /* Equation 8 */
  private static double siteTermLinear(Coefficients c, double vs30) {
    double Flin;
    if (vs30 <= V1) {
      Flin = c.s1 * log(vs30 / V1) + c.s2 * log(V1 / VREF);
    } else if (vs30 > c.v2) {
      Flin = c.s2 * log(c.v2 / VREF);
    } else {
      Flin = c.s2 * log(vs30 / VREF);
    }
    return Flin;
  }

  /* Equation 9 */
  private static double siteTermNonLinear(Coefficients c, double pgaRef, double vs30) {
    double f3 = 0.05;
    double f2 = c.f4 * (exp(c.f5 * (min(vs30, 760) - 200)) - exp(c.f5 * (760 - 200)));
    return f2 * log((pgaRef + f3) / f3); // f1 = 0
  }

  /* Equation 11, 12, 13 */
  private static double siteBasinTerm(
      Coefficients c,
      double vs30,
      double z2p5,
      boolean slab,
      boolean basin,
      boolean m9) {

    if (Double.isNaN(z2p5)) {
      return 0.0;
    }

    double z2p5m = z2p5 * 1000.0;

    // Cascadia coefficients
    double θ0 = 3.94;
    double θ1 = -0.42;
    double νμ = 200.0;
    double νσ = 0.2;

    double lnμz2p5 = log(10) *
        (θ1 * (1 + Maths.erf((log10(vs30) - log10(νμ)) / νσ / sqrt(2))) + θ0);
    double δz2p5 = log(z2p5m) - lnμz2p5;

    double Fb = calcFb(c, δz2p5);

    if (m9) { // basin is implicitely true
      if (!slab && z2p5 > 6.0 && (c.imt.isSA() && c.imt.period() > 1.9)) {
        double δz2p5adj = log(z2p5m) - LN_Z2P5_REF_PARKER;
        double m9adj = log(2.0) - calcFb(c, δz2p5adj);
        Fb += m9adj;
      }
    }

    if (basin) {
      Fb *= GmmUtils.deltaZ25scale(c.imt, z2p5);
    }

    return Fb;
  }

  private static double calcFb(Coefficients c, double δz2p5) {
    if (δz2p5 <= (c.c_e1 / c.c_e3)) {
      return c.c_e1;
    } else if (δz2p5 >= (c.c_e2 / c.c_e3)) {
      return c.c_e2;
    }
    return c.c_e3 * δz2p5;
  }

  /* Eqsuation 18, 19, 20 */
  private static double phiTotal(Coefficients c, double rRup, double vs30) {

    double Δvar;
    double rPrime = max(ΦR1, min(ΦR2, rRup));
    if (vs30 <= ΦV1) {
      Δvar = c.ϕ2v * (log(ΦR2 / rPrime) / log(ΦR2 / ΦR1));
    } else if (vs30 >= ΦV2) {
      Δvar = 0;
    } else {
      Δvar = c.ϕ2v *
          (log(ΦV2 / vs30) / log(ΦV2 / ΦV1)) *
          (log(ΦR2 / rPrime) / log(ΦR2 / ΦR1));
    }

    double ϕ2;
    if (rRup <= ΦR1) {
      ϕ2 = c.ϕ21;
    } else if (rRup >= ΦR2) {
      ϕ2 = c.ϕ22;
    } else {
      ϕ2 = (c.ϕ22 - c.ϕ21) / log(ΦR2 / ΦR1) * log(rRup / ΦR1) + c.ϕ21;
    }

    return sqrt(Δvar + ϕ2);
  }

  /* Equation 27 */
  private static double calcEpistemic(Coefficients c) {
    double σε;
    if (c.imt == Imt.PGA || c.imt == Imt.PGV || c.imt.period() < c.t1) {
      σε = c.σε1;
    } else if (c.imt.period() < c.t2) {
      σε = c.σε1 - (c.σε1 - c.σε2) * log(c.imt.period() / c.t1) / log(c.t2 / c.t1);
    } else {
      σε = c.σε2;
    }
    return σε;
  }

  private static abstract class Global extends ParkerEtAl_2020 {
    static final SubductionZone ZONE = GLOBAL;
    static final String NAME = ParkerEtAl_2020.NAME + " : " + ZONE;

    Global(Imt imt) {
      super(imt, ZONE);
    }
  }

  private static abstract class Cascadia extends ParkerEtAl_2020 {
    static final SubductionZone ZONE = CASCADIA;
    static final String NAME = ParkerEtAl_2020.NAME + " : " + ZONE;

    Cascadia(Imt imt) {
      super(imt, ZONE);
    }
  }

  private static abstract class Alaska extends ParkerEtAl_2020 {
    static final SubductionZone ZONE = ALASKA;
    static final String NAME = ParkerEtAl_2020.NAME + " : " + ZONE;

    Alaska(Imt imt) {
      super(imt, ZONE);
    }
  }

  static final class GlobalInterface extends Global {
    static final String NAME = Global.NAME + " : " + INTERFACE;

    GlobalInterface(Imt imt) {
      super(imt);
    }
  }

  static final class GlobalInterfaceAkAdjusted extends Global {
    static final String NAME = GlobalInterface.NAME + " (Alaska Adjusted)";

    GlobalInterfaceAkAdjusted(Imt imt) {
      super(imt);
    }

    @Override
    boolean akBiasAdjusted() {
      return true;
    }
  }

  static final class GlobalSlab extends Global {
    static final String NAME = Global.NAME + " : " + SLAB;

    GlobalSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static class CascadiaInterface extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + INTERFACE;

    CascadiaInterface(Imt imt) {
      super(imt);
    }
  }

  static class CascadiaInterfaceBasin extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " : Basin";

    CascadiaInterfaceBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaInterfaceBasinM9 extends CascadiaInterfaceBasin {
    static final String NAME = CascadiaInterfaceBasin.NAME + " : M9";

    CascadiaInterfaceBasinM9(Imt imt) {
      super(imt);
    }

    @Override
    boolean m9() {
      return true;
    }
  }

  static class CascadiaSlab extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + SLAB;

    CascadiaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static final class CascadiaSlabBasin extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " : Basin";

    CascadiaSlabBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class AlaskaInterface extends Alaska {
    static final String NAME = Alaska.NAME + " : " + INTERFACE;

    AlaskaInterface(Imt imt) {
      super(imt);
    }
  }

  static final class AlaskaSlab extends Alaska {
    static final String NAME = Alaska.NAME + " : " + SLAB;

    AlaskaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }
}
